// -*- c++ -*-
// Attribute.cc
//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
// $Log: Attribute.cc,v $
// Revision 0.4  1997/12/23 22:30:43  sellden
// Fixed casts in creation of Attribute::implicit[] to suppress warnings
//
// Revision 0.3  1997/02/16 09:06:26  sellden
// Removed dictionary constructors which take a hashing function
// pointer.  Hashing functionality is now determined by
// overriding the class HashActions.
//
// Revision 0.2  1996/12/11 18:34:02  barnaby
// Changed NULL to be 0 problems with void*
//
// Revision 0.1  1996/12/02 20:06:39  rcsg
// Initial Port Revision
//
//
//

#include "Attribute.h"
#include "Mercator.h"
#include "Error.h"

extern FILE *attribin;

#ifdef USONIX //blsx: outmoded constructor
mDictValPtr<mString, Attribute> *Attribute::classAttributeDictionary =
  new mDictValPtr<mString, Attribute>();

mDictValPtr<mString, Attribute> *Attribute::appAttributeDictionary =
  new mDictValPtr<mString, Attribute>();
#else
static unsigned StringHasher(const mString& str) { return str.hash(); }

mDictValPtr<mString, Attribute> *Attribute::classAttributeDictionary =
  new mDictValPtr<mString, Attribute>(StringHasher);

mDictValPtr<mString, Attribute> *Attribute::appAttributeDictionary =
  new mDictValPtr<mString, Attribute>(StringHasher);
#endif

int Attribute::validDicts = 0;

extern "C" {
    
void
ProcessAttributeLine(const char *name, int ftype, const char *defaultname,
		     int cltype)
{
    Attribute *a;
    int x;

    a = new Attribute(name, defaultname, (FieldType) ftype,
		      (Attribute::COType) cltype);

    a->InstallAttribute();
}

} // extern "C"

int
Attribute::ParseAttribFile(const char *filename)
{
    FILE *attribFile = NULL;

    // empty dictionary if necessary
    // this allows use to call ParseAttribFile anytime we want to re-source
    // the attrib file
    if (Attribute::validDicts)
	Attribute::EmptyDicts();


    // install implicit attributes first
    Attribute::InstallImplicitAttributes();
    
    if (filename == NULL) {
	Error(EL_ERROR, "Attribute::ParseAttribFile",
	      "No filename specified.", NULL);
	return 0;
    }
    
    if ((attribFile = fopen(filename, "r")) == NULL) {
	Error(EL_ERROR, "Attribute::ParseAttribFile", "couldn't open file ",
	      filename, ": ", Perror(), NULL);
	return 0;
    }
    
    // attribin is the yyin for the attribute parser.
    attribin = attribFile;
    
    // attribparse() is the yyparse for the attriblate parser.
    if (attribparse() != 0) {
	Error(EL_ERROR, "Attribute::ParseAttribFile", "Parser failed for ",
	      filename, NULL);
        fclose(attribin);
	return 0;
    }
    
    // close the file
    Attribute::validDicts = 1;
    fclose(attribin);
    return 1;

}

int
Attribute::EmptyDicts()
{
    Attribute::appAttributeDictionary->kill_all();
    Attribute::classAttributeDictionary->kill_all();

    delete Attribute::appAttributeDictionary;
    delete Attribute::classAttributeDictionary;

#ifdef USONIX //blsx: outmoded constructor
    Attribute::appAttributeDictionary =
	new mDictValPtr<mString, Attribute>();
    Attribute::classAttributeDictionary =
	new mDictValPtr<mString, Attribute>();
#else
    Attribute::appAttributeDictionary =
	new mDictValPtr<mString, Attribute>(StringHasher);
    Attribute::classAttributeDictionary =
	new mDictValPtr<mString, Attribute>(StringHasher);
#endif

    Attribute::validDicts = 0;

    return 1;
}

void
Attribute::FreeAttrib()
{
    Attribute::EmptyDicts();

    delete Attribute::appAttributeDictionary;
    delete Attribute::classAttributeDictionary;

    Attribute::validDicts = 0;
}

// constructor
Attribute::Attribute(const char *n, const char *ref_name,
		     FieldType t, COType cotype) : 
    name(Strdup(n)), tag(t),reference_name(Strdup(ref_name)), ctype(cotype)
{
}

Attribute::~Attribute()
{
    delete [] name;
    delete [] reference_name;

    return;
}

int
Attribute::InstallAttribute()
{
    Attribute *a;

    /* lookup value to see if it already exists.  Delete it if present */
    if ((ctype == CLASS) && (a = classAttributeDictionary->lookup(name)))
	delete a;

    if ((ctype == APP) && (a = appAttributeDictionary->lookup(this->name)))
	delete a;

    if (ctype == CLASS)
	classAttributeDictionary->insert(name, this);
    else if (ctype == APP)
	appAttributeDictionary->insert(this->name, this);
    else {
	Error(EL_ERROR, "Attribute::InstallAttribute", "invalid CType", NULL);
	return -1;
    }

    return 0; // zero indicates success
   
}

Attribute *
Attribute::GetClassAttributeByName(const char *n)
{
    return classAttributeDictionary->lookup(n);
}

Attribute *
Attribute::GetAppAttributeByName(const char *n)
{
    return appAttributeDictionary->lookup(n);
}

void
Attribute::InstallImplicitAttributes()
{
    int count;

    count = 0;
    while (implicit[count].name) {
	ProcessAttributeLine(implicit[count].name, implicit[count].tag,
			     implicit[count].refname, implicit[count].ctype);
	count++;
    }
}


Attribute::AttribStruct Attribute::implicit[] = {
    "navigable",      BOOLEAN,      "defaultClassNavigable",         CLASS,
    "shell",          BOOLEAN,      "defaultClassShell",             CLASS,
    "allowTextMode",  BOOLEAN,      "defaultClassAllowTextMode",     CLASS,
    "sensitive",      BOOLEAN,      "defaultClassSensitive",	     CLASS,
    "looped",         BOOLEAN,      "defaultClassLooped",	     CLASS,
    "mappedWhenManaged", BOOLEAN,   "defaultClassMappedWhenManaged", CLASS,
    "allowEscape",    BOOLEAN,      "defaultClassAllowEscape",       CLASS,
    "editable",       BOOLEAN,      "defaultClassEditable",          CLASS,

    "volume",         INTEGER,      "defaultClassVolume",            CLASS,
    "muffle",         INTEGER,      "defaultClassMuffle",            CLASS,
    "rate",           INTEGER,      "defaultClassRate",              CLASS,
    "delay",          INTEGER,      "defaultClassDelay",             CLASS,

    "leftMargin",     INTEGER,      "defaultClassLeftMargin",        CLASS,
    "rightMargin",    INTEGER,      "defaultClassRightMargin",       CLASS,
    "topMargin",      INTEGER,      "defaultClassTopMargin",         CLASS,
    "bottomMargin",   INTEGER,      "defaultClassBottomMargin",      CLASS,

    "sound",          STRING,       "defaultClassSound",             CLASS,
    "speakOnInfo",    STRING,       "defaultClassSpeakOnInfo",       CLASS,
    "speakOnEnter",   STRING,       "defaultClassSpeakOnEnter",      CLASS,

    "unsafeResources", STRINGLIST,  "defaultClassUnsafeResources",   CLASS,

    "creation",	      INTEGER,	    "defaultClassCreation",	     CLASS,
    "destruction",    INTEGER,	    "defaultClassDestruction",	     CLASS,
    "mapped",	      INTEGER,	    "defaultClassMapped",	     CLASS,
    "unmapped",	      INTEGER,	    "defaultClassUnmapped",	     CLASS,
    "managed",	      INTEGER,	    "defaultClassManaged",	     CLASS,
    "unmanaged",      INTEGER,	    "defaultClassUnmanaged",	     CLASS,
    "realized",	      INTEGER,	    "defaultClassRealized",	     CLASS,
    "unrealized",     INTEGER,	    "defaultClassUnrealized",	     CLASS,

    // Default filters to load for each ScreenReader
    "srFilters",      STRINGLIST,   "defaultClassSrFilters",	     CLASS,
    // Characters which will be treated as word delimiters
    "srDelimitChars", STRING,	    "defaultClassSrDelimitChars",    CLASS,
    // Characters which will be treated as sentence terminators
    "srTerminalChars",STRING,	    "defaultClassSrTerminalChars",   CLASS,
    // Startup/shutdown sounds.  These aren't used yet.
    "readySound",     STRING,       "defaultAppReadySound",          APP,
    "shutdownSound",  STRING,       "defaultAppShutdownSound",       APP,
    // The the application draw its cursor as a block?
    "blockCursor",    BOOLEAN,	    "defaultAppBlockCursor",	     APP,
    // Give feedback whenever the app cursor moves?
    "cursorFeedback", BOOLEAN,      "defaultAppCursorFeedback",      APP,
    // What sound to play
    "cursorSound",    STRING,       "defaultAppCursorSound",         APP,

#ifdef USONIX	
    NULL, BOOLEAN, NULL, CLASS
#else
    0, 0, 0, 0
#endif
};
