//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
// $Log: ProcessManager.h,v $
// Revision 0.4  1997/12/23 22:30:35  sellden
// Added support for X to perform polling for FDInterest instances
//
// Revision 0.3  1996/12/11 18:41:49  barnaby
// Signal action handlers take an int under Linux added linux definition
//
// Revision 0.2  1996/12/06 15:15:02  sellden
// Replaced template forward declarations with headers
//
// Revision 0.1  1996/12/02 20:06:39  rcsg
// Initial Port Revision
//
//
// The ProcessManager provides a way of tracking subprocesses.  Every
// program which spawns subprocesses should instantiate a ProcessMgr
// to deal with them.
//
// The ProcessManager is a subclass of FDInterest.  It creates a pipe to
// itself which it uses to handle process-related activity in a signal-safe
// manner.  The signal handler for SIGCHLD writes a byte onto the
// ProcessManager pipe which causes the appropriate action to fire.
//
// ProcessManager provides the following services -
//	- Wrapper functions to start, stop, and signal processes.
//	- SIGCHLD handling for all subprocesses.
//	- Automatic shutdown of subprocesses on destruction (optional).
//	- Access to subprocess command lines via pid.
//
// You can only create one ProcessManager in a given address space.
//
//	keith edwards	14jul95
//

#ifndef _PROCESS_MGR_H_
#define _PROCESS_MGR_H_

#include <sys/types.h>
#include <signal.h>

#ifdef USONIX
#define String	XtString
#include <X11/Intrinsic.h>
#undef String
#endif

#include "FDInterest.h"
#include <mDictValPtr.h>

#ifndef TRUE
#	define	TRUE 1
#endif

#ifndef FALSE
#	define	FALSE 0
#endif

class ProcessManager : public FDInterest {
public:
    typedef int (*UnexpectedDeathProc)(pid_t);
    typedef int (*BeforeDeathProc)(pid_t);
    typedef int (*AfterDeathProc)(pid_t);
		 
private:
    // Write on 0, read on 1
    static int		pipeFds[2];

    // Have we created an instance yet?
    static int		initialized;

    // Have we started the shutdown process?
    int			shuttingDown;

#ifdef USONIX
    // Returned by XtAppAddInput
    XtInputId inputID;
#endif

    // signal handlers
    static struct sigaction	action, oldAction;

    typedef struct _ProcessNode {
	int argc;
	char **argv;
	char *basename;
	int killOnShutdown;
	int useClientEnvironment;
	int flaggedForKill;
	UnexpectedDeathProc unexpected;
	BeforeDeathProc avant;
	AfterDeathProc apres;

	~_ProcessNode();
    } ProcessNode;

    mDictValPtr<pid_t, ProcessNode>	*processes;

    // Disallow default operations
    ProcessManager(const ProcessManager&);
    ProcessManager&	operator=(const ProcessManager&);
#if	defined(SUNOS_5_5) || defined(linux)	// TB: Linux signals have arg
    static void		Reaper(int);
#else
    static void		Reaper();
#endif /* SUNOS_5_5 */
    
public:
    ProcessManager();
    virtual ~ProcessManager();

    //
    // This is the only interface to ProcessManager that most code will
    // see.  The args are used as follows:
    //
    //    argc, argv, envp - passed to exec.
    //	killOnShutdown - should the process be terminated when
    //		Mercator goes away?
    //   unexpected - Optional procedure called whenever the child
    //		unexpectedly terminates.  Can be used to restart, if
    //		desired.
    //	avant - Optional procedure called whenever the child is about
    //		to be killed by Mercator.
    //	apres - Optional procedure called after the child has been terminated
    //		by Mercator.
    //
    // StartProcess returns the pid of the child process.
    //
    pid_t		StartProcess(int argc, char * const argv[],
				     char * const envp[] = NULL,
				     int killOnShutdown = TRUE,
				     UnexpectedDeathProc unexpected = NULL,
				     BeforeDeathProc avant = NULL,
				     AfterDeathProc apres = NULL);
    
    int			StopProcess(pid_t);
    int			SuspendProcess(pid_t);
    int			ResumeProcess(pid_t);
    int			SignalProcess(pid_t, int);

#ifdef USONIX
  static void HandleXtActivity(XtPointer client_data,
		int	*source, 	      XtInputId	*id);
XtInputId InputID()
  {return inputID;}
void InputID(XtInputId newID)
  {inputID=newID;}
#endif

    virtual int		HandleActivity(int fd, short mask);
    virtual void	Shutdown(int fd);
};

#endif /* _PROCESS_MGR_H_ */
