// This may look like C code, but it is really -*- C++ -*-
//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
//
#ifndef MERCATOR_LIST_H
#define MERCATOR_LIST_H

// $Author: sellden $
// $Source: /usr/mercator/newsource/7.0/RCS/mList.h,v $
// $Revision: 0.6 $
// $Date: 1998/05/21 23:01:07 $

// $Log: mList.h,v $
// Revision 0.6  1998/05/21 23:01:07  sellden
// Added forward decl for iterator template
//
// Revision 0.5  1997/12/23 22:30:45  sellden
// Added 'const' in del_item()
//
// Revision 0.4  1997/02/01 23:20:20  sellden
// Completed migration to OTC_Lists and OTC_Iterators.
//
// Revision 0.2  1996/12/12 14:31:33  zimba
// ose implementation of mList
//
//Revision 1.4  1995/07/28  21:23:35  spine
//Added copyright notice.
//
//Revision 1.3  1995/07/24  17:44:26  beth
//Ian's Version
//
//Revision 1.1  1995/05/25  14:48:43  iansmith
//Initial revision
//

// We implement the mList with a Rogue Wave List right now

#include <OTC/collctn/list.hh>
#include <OTC/collctn/iterator.hh>

#ifdef USONIX // BLS: EGCS compiler was complaining
template<class T> class mListIterator;
#endif

//
// This is the closure object that you pass to the iterate function.
// Subclass the action function to implement your behavior;
// if you want startup and closedown help you can implement those
// as well.
//
template <class T> class mListClosure {
public:
    // use this to setup your closure
    virtual void startup(void *clientData) {};
    // you'll still get a closedown if you short circuit
    //  out of the action routine
    virtual void closedown() {};
    // if the action routine returns 0 the closure stops
    // processing
    // if you don't subclass, we end up only calling the action routine once
    virtual int action(T *) {return 0;};
};

// 
// Linked List for Mercator
// You can ONLY put pointers in this data structure!
//
// You MUST define the following:
// operator==
// operator<
//
template <class T> class mList 
{
  // The implementation
  
  OTC_List<T*> *impl;
  friend class mListIterator<T>;
public:
  // add to the end
  void append(const T *v){impl->addLast((T* const)v);};

  // search for a value... returns 0 if not found
  T * search(const T* val) const{
    T* found=NULL;//assume not found
    for(int i=0;i<impl->population();i++){
      if(*val == *(impl->item(i))){
	found=(T*)val;
	break;
      }
    }
    return found;
  };

  // remove an item and call its destructor
  // This is noop if its not present
  void destroy_item(T *val){
    for(int i=0;i<impl->population();i++){
      if(*val == *(impl->item(i))){
	impl->removeItem(i);
	delete val;
	break;
      }
    }
  };

  // remove the item but return a pointer to it...
  // comparison is with the == operator on T
  const T* del_item(const T *val){
    const T *result=(const T*)NULL;

    for(int i=0;i<impl->population();i++){
      if(*val == *(impl->item(i))){
	impl->removeItem(i);
	result = val;
	break;
      }
    }
    return (const T*)result;
  };

  // if you want to iterate, subclass the mListClosure and
  // pass it in here... the client data is passed to the startup routine
  void walk(mListClosure<T> *closure,void *clientData) const{
    //issue the startup
    closure->startup(clientData);
    //go through the list
    for(int i=0;i<impl->population();i++){
      if(closure->action(impl->item(i)) == 0)
	break;

    }
  };

  // constructor (empty list)
  mList(){
    impl=new OTC_List<T*>;
  };

  // copy constructor.. this will result in *shallow* copy with pointers
  // shared between lists
  mList(const mList &x){
    impl=new OTC_List<T*>(*x.impl,OTCLIB_SHALLOWCOPY);
  };

  // destructor... this just deletes the data structure, doesn't touch the objects
  virtual ~mList(){delete impl;};
  // assignment is a shallow copy of the lists... you get two differents lists
  // but the nodes in them are the same
  mList<T> & operator=(const mList<T> &x){
    *impl=*(x.impl);
    return *this;
  };

  // sometimes you want information about ordering...
  // get the head of the list
  T * head() const{return(impl->first());};

  // get the tail of the list
  T * tail() const{return(impl->last());};

  // get the entry right before or after... 
  // return 0 if they don't have the entry asked for or
  // succ/pred would be off the end/beginning of the list
  // both of these test for equality using op == on T
  T * predecessor(const T *val) const{
    T* result=NULL;
    //use two indexes h and i
    //h=i-1
    for(int h=-1,i=0;i<impl->population();i++,h++){
      if(*val == *(impl->item(i))){
	if(h>=0&&h<impl->population())
	  result=impl->item(h);

	break;
      }
    }
    return result;
  };

  T * successor(const T *val) const{
    T* result=NULL;
    //use two indexes h and i
    //h=i-1
    for(int h=1,i=0;i<impl->population();i++,h++){
      if(*val == *(impl->item(i))){
	if(h>=0&&h<impl->population())
	  result=impl->item(h);

	break;
      }
    }
    return result;
  };

  // how big is the list?
  unsigned entries() const{return(impl->population());};

  // is it empty?
  int empty() const{return(impl->isEmpty());};
};

// OTC iterators are only undefined
// a) If they are constructed as a null iterator - that is,
//    not associated with any list
//
//  b)  They iterate off the end of a list.
//
//  c) They are assigned to an iterator that is invalid.
//
// You should not  remove elements from the list during iteration.  
template <class T> class mListIterator {
  OTC_Iterator<T*> impl;

public:

  // construct this from an mList
  mListIterator(const mList<T> &x)
  {       impl = x.impl->items();};

  // copy constructor
  mListIterator(const mListIterator &x){impl=x.impl;};

  // destructor
  virtual ~mListIterator(){};

  // next() was being used to initialize iterators plus iterate.
  // OTC iterators do not need to be initialized unless they were
  // created without associated lists.  We make sure this doesn't 
  // happen by providing no such constructors.
  T * next(){
    if(impl.isValid()){
      //advance to the next item
      impl.next();
      if(impl.isValid())
	return impl.item();
    else
    return (T*)NULL;
    }
  };

  // Reset iterator to point to top of list, iter is still valid after.
  void reset(){impl.reset();};

  // Return current item
  T* current() { return impl.item();};

  // operator = 
  //  mListIterator<T*> & operator=(const mListIterator<T*> &iter){impl = iter;};
};

#endif /* MERCATOR_LIST_H */





