/************************************************************************/
/*									*/
/*   Copyright 1996 by Digital Equipment Corporation, Maynard, Mass.	*/
/*									*/
/*                        All Rights Reserved				*/
/*								        */
/* Permission to use, copy, modify, and distribute this software and    */
/* its documentation for any purpose and without fee is hereby granted,	*/
/* provided that the above copyright notice appear in all copies and 	*/
/* that both that copyright notice and this permission notice appear in */
/* supporting documentation, and that the name Digital not be used in	*/
/* advertising or publicity pertaining to distribution of the software	*/
/* without specific, written prior permission.				*/
/*									*/
/* DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,	*/
/* INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN  */
/* NO EVENT SHALL DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR	*/
/* CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS	*/
/* OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, 		*/
/* NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 		*/
/* CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.		*/
/*								        */
/************************************************************************/
/*
* $Log: RapClnt.c,v $
* Revision 0.1  1997/09/02 20:49:03  rcsg
* Initial revision
*
*/

/************************************************************************/
/*									*/
/* RapClnt.c								*/
/*									*/
/*      RAP is a simple ICE subprotocol that uses Xt hooks to share	*/
/*	share client information with external agents.	 This file	*/
/*	handles general RAP information for clients.	 		*/
/*									*/
/************************************************************************/
#include <X11/Xmd.h>
#include <X11/Xlib.h>
#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>

#include <X11/ICE/ICElib.h>
#include <X11/ICE/ICEmsg.h>
#include <X11/ICE/ICEproto.h>
#include "Xtea.h"
#include "Rap.h"
#include "RapProto.h"
#include "RapClnt.h"

extern void
RapClntFlushProc(Display 	*dpy, 
		 XExtCodes	*ext, 
		 char 		*data, 
		 long 		len);

static int major_opcode = 0;
XExtCodes *rap_ext_codes = NULL;

/************************************************************************/
/*									*/
/* ErrorHandler								*/
/*									*/
/* 	Called when an unexpected ICE protocol error is encountered.	*/
/*	[[[WDW - Dying agents should be handled differently.]]]		*/
/*									*/
/************************************************************************/
static void 
ErrorHandler(IceConn 		ice_conn,
	     Bool		swap,
	     int		minor_opcode,
	     unsigned long	sequence_num,
	     int		error_class,
	     int		severity,
	     IcePointer		values)
{
    ErrorF(__FILE__,__LINE__,"ErrorHandler",
	   "minor_opcode=%d,sequence_num=%d,error_class=%d,severity=%d\n",
	   minor_opcode,sequence_num,error_class,severity);
    
} /* ErrorHandler */


/************************************************************************/
/*									*/
/* IOErrorHandler							*/
/*									*/
/* 	Called when an unexpected ICE protocol error is encountered.	*/
/*	[[[WDW - Dying agents should be handled differently.]]]		*/
/*									*/
/************************************************************************/
static void 
IOErrorHandler(IceConn ice_conn)
{
    ErrorF(__FILE__,__LINE__,"IOErrorHandler","ice_conn=%x\n",ice_conn);

} /* IOErrorHandler */


/************************************************************************/
/*									*/
/* IOErrorProc								*/
/*									*/
/* 	Called if the ICE connection unexpectedly breaks.  This routine	*/
/*	is established as the IO error handler during the call to	*/
/*	IceRegisterForProtocolSetup.					*/
/*									*/
/************************************************************************/
static void 
IOErrorProc(IceConn ice_conn)
{
    ErrorF(__FILE__,__LINE__,"IOErrorProc","ice_conn=%x\n",ice_conn);
    IceCloseConnection(ice_conn);

} /* IOErrorProc */


/************************************************************************/
/*									*/
/* RapClntAddAgent							*/
/*									*/
/*	Add the agent to the list of known agents.			*/
/*									*/
/************************************************************************/
void
RapClntAddAgent(Display  *dpy,
		RapInfo  ri)
{
    XEDataObject	ext_object;
    XExtData 		*ext_data;

    ext_object.display = dpy;
    if (rap_ext_codes == NULL) {
	rap_ext_codes = XAddExtension(dpy);
	ext_data = (XExtData *) XtMalloc(sizeof(XExtData));
	memset(ext_data,'\0',sizeof(XExtData));
	ext_data->number = rap_ext_codes->extension;
	XAddToExtensionList(XEHeadOfExtensionList(ext_object),
			    ext_data);

	/* Set up a hook that will determine the requests when the
	 * display is flushed.
	 *
	 * [[[ WDW - Doing this now is pretty inefficient for folks
	 *     who do not want to get requests. ]]]
	 */
	XESetBeforeFlush(dpy,rap_ext_codes->extension,RapClntFlushProc);
    }
    else
	ext_data = XFindOnExtensionList(XEHeadOfExtensionList(ext_object),
					rap_ext_codes->extension);

    /* Add it to the head of the table.
     */
    ri->next = (RapInfo) ext_data->private_data;    
    ext_data->private_data = (XPointer) ri;

} /* RapClntAddAgent */


/************************************************************************/
/*									*/
/* RemoveAgent								*/
/*									*/
/*	Remove the agent from the list of known agents.			*/
/*									*/
/************************************************************************/
static void
RemoveAgent(XExtData *ext_data,
	    RapInfo  ri)
{
    RapInfo current,prev;

    current = (RapInfo) ext_data->private_data;

    while (current != NULL) {
	if ((current->agent_window == ri->agent_window) &&
	    (current->ice_conn == ri->ice_conn)) {
	    if (prev)
		prev->next = current->next;
	    else
		ext_data->private_data = (XPointer) current->next;
	}
	prev = current;
	current = current->next;
    }

} /* RemoveAgent */


/************************************************************************/
/*									*/
/* ProcessIceMsg							*/
/*									*/
/* 	Responsible for processing the set of messages that can	        */
/*	be received from ICE.  This routine is set up as the processor	*/
/*	via a call to XtAppAddInput by RapClntInit, the routine 	*/
/*	that establishes the ICE connection with the agent.		*/
/*									*/
/************************************************************************/
static void 
ProcessIceMsg(XtPointer client_data, 
	      int	*source, 
	      XtInputId	*id)
{
    IceProcessMessagesStatus	status;
    RapInfo			ri = (RapInfo)client_data;
    XEDataObject		ext_object;
    
    /* IceProcessMessages ends up calling ProcessMsg
     * because it is the message handler for the RAP protocol.
     *
     * [[[WDW - I'm not quite sure about what the reply_ready
     *		stuff does, and if we should be taking advantage
     *		of it somehow.]]]
     */
    status = IceProcessMessages(ri->ice_conn,
				NULL,	/* reply_ready_ret */
				NULL);	/* reply_ready	   */
    
    /* Check the status to see if there were any problems.
     */
    if (status != IceProcessMessagesSuccess) {
	switch(status) {
	    case IceProcessMessagesIOError: {
		ErrorF(__FILE__,__LINE__,"ProcessIceMsg",
		       "IceProcessMessagesIOError.  Closing connection\n",
		       NULL);
		IceCloseConnection(ri->ice_conn);
		XtRemoveInput(*id);
		break;
	    }
	    case IceProcessMessagesConnectionClosed: {
#ifdef DEBUG
		InfoF(__FILE__,__LINE__,"ProcessIceMsg",
		      "IceProcessMessagesConnectionClosed\n");
#endif
		break;
	    }
	    default: {
		ErrorF(__FILE__,__LINE__,"ProcessIceMsg",
		       "Unknown status %d\n",status);
		break;
	    }
	}
	RapClntRemoveCallbacks(ri);
	ext_object.display = XtDisplay(ri->top_level);
	RemoveAgent(XFindOnExtensionList(XEHeadOfExtensionList(ext_object),
					 rap_ext_codes->extension),
		    ri);
	XtFree((char *)ri);
	ErrorF(__FILE__,__LINE__,"ProcessIceMsg",
	       "Freed RapInfo data.\n");
    }

} /* ProcessIceMsg */

/************************************************************************/
/*									*/
/* RapClntDispatchMsg							*/
/*									*/
/* 	Responsible for handling/dispatching the set of messages that 	*/
/*	can be received from the protocol.				*/
/*									*/
/************************************************************************/
void
RapClntDispatchMsg(RapInfo	ri,
		   int		opcode,
		   CARD16	seq,
		   char		*pdata)
{
    switch(opcode) {
	/* The RapHelloRequest opcode is used by the agent just after
	 * it accepts the ICE connection.  It expects the client to
	 * respond with a RapHelloMessage that contains the ID of the
	 * client's top level window.  This allows the agent to do the
	 * bookkeeping necessary to handle multiple clients.
	 */
	case RapHelloRequest: {	    
	    RapClntHelloReply(ri);
	    break;
	}

	case RapQueryTreeRequest: {	    
	    RapClntQueryTreeReply(ri,seq,False);
	    break;
	}

	case RapFullQueryTreeRequest: {	    
	    RapClntQueryTreeReply(ri,seq,True);
	    break;
	}

	case RapGetResourcesRequest: {	    
	    RapClntHandleGetResourcesRequest(ri,seq,pdata);
	    break;
	}

#ifdef GC_SUPPORT
	case RapGetGCValuesRequest: {	    
	    RapClntHandleGetGCValuesRequest(ri,seq,pdata);
	    break;
	}
#endif

	case RapGetValuesRequest: {	    
	    RapClntHandleGetValuesRequest(ri,seq,pdata);
	    break;
	}

	case RapSetValuesRequest: {	    
	    RapClntHandleSetValuesRequest(ri,seq,pdata);
	    break;
	}

	case RapGetHooksRequest: {	    
	    RapClntGetHooksReply(ri,seq);
	    break;
	}

	case RapSelectHooksRequest: {	    
	    RapClntHandleSelectHooksRequest(ri,seq,pdata);
	    break;
	}

	case RapGetEventsRequest: {	    
	    RapClntGetEventsReply(ri,seq);
	    break;
	}

	case RapSelectEventsRequest: {	    
	    RapClntHandleSelectEventsRequest(ri,seq,pdata);
	    break;
	}

	case RapGetRequestsRequest: {	    
	    RapClntGetRequestsReply(ri,seq);
	    break;
	}

	case RapSelectRequestsRequest: {	    
	    RapClntHandleSelectRequestsRequest(ri,seq,pdata);
	    break;
	}

	case RapLocateObjectRequest: {	    
	    RapClntHandleLocateObjectRequest(ri,seq,pdata);
	    break;
	}

	default: {
	    ErrorF(__FILE__,__LINE__,"ProcessMsg",
		   "unknown opcode - %d\n",opcode);
	    break;
	}
    } /* switch(opcode) */

} /* RapClntDispatchMsg */

/************************************************************************/
/*									*/
/* ProcessMsg								*/
/*									*/
/*	Responsible for processing RAP protocol messages.  This routine	*/
/*	was assigned as the process message proc during the call to	*/
/*	IceRegisterForProtocolSetup by RapClntInit.  It is called 	*/
/*	as a result of ProcessIceMsgProc calling IceProcessMessages.	*/
/*									*/
/*	The client_data parameter is set to the widget that was passed	*/
/*	to RapClntInit.							*/
/*									*/
/************************************************************************/
static void 
ProcessMsg(IceConn		ice_conn,
	   IcePointer 		client_data,
	   int			opcode,
	   unsigned long	length,
	   Bool			swap,
	   IceReplyWaitInfo	*reply_wait,
	   Bool			*reply_ready_ret)
{
    RapInfo 		ri = (RapInfo)client_data;
    RapMessageHeader 	*header;
    char		*pdata;
    
    IceReadCompleteMessage(ice_conn,
			   SIZEOF(RapMessageHeader),
			   RapMessageHeader,
			   header,
			   pdata);
    if (!IceValidIO(ice_conn)) {
	ErrorF(__FILE__,__LINE__,"ProcessMsg",
	       "IceValidIO returned False on connection.\n");
    }

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"ProcessMsg",": ");
#endif

    RapClntDispatchMsg(ri,opcode,header->sequenceNumber,pdata);
        
    IceDisposeCompleteMessage(ice_conn, pdata);

} /* ProcessMsg */

/************************************************************************/
/*									*/
/* RapClntInit								*/
/*									*/
/*	Initiate the ICE connection with the agent.  This routine is	*/
/*	called by XteaCheckMessages as a result of the client 		*/
/*	registering the RAP protocol with a call to XteaSetupClient.	*/
/*									*/
/*	In addition to initiating the connection, this routine also 	*/
/*	sets up the callbacks that work with the external agent hooks	*/
/*	built into Xt.  The client data used for the hooks is the ICE 	*/
/*	connection ID.	This allows clients to keep track of multiple 	*/
/*	agents.								*/
/*									*/
/************************************************************************/
void 
RapClntInit(Widget	       	w,
	    XClientMessageEvent *cme,
	    char 		*network_ids,
	    XtPointer		client_data)
{
    IcePoVersionRec		protocol_versions[1];
    IceConn			ice_conn;
    char			error_str[256];
    char			*connect_str;
    IceProtocolSetupStatus	status;
    int				major_version,minor_version;
    char			*vendor_str,*release_str;
    Display			*dpy = XtDisplay(w);
    RapInfo			rap_info;
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapClntInit","\n");
#endif

    /* If this is the first time this routine has been called,
     * do the necessary work to let ICE know about Rap and that
     * this client will be able to do a protocol setup for Rap.
     */
    if (!major_opcode) {
	protocol_versions[0].major_version    = RapMajorVersion;
	protocol_versions[0].minor_version    = RapMinorVersion;
	protocol_versions[0].process_msg_proc = ProcessMsg;
    
	major_opcode = IceRegisterForProtocolSetup(RapName,
						   RapVendor,
						   RapRelease,
						   1,protocol_versions,
						   0,NULL,NULL,
						   IOErrorProc);

	if (major_opcode == -1) {
	    ErrorF(__FILE__,__LINE__,"RapClntInit",
		   "IceRegisterForProtocolSetup failed.\n");
	    XteaSendFailure(dpy,cme,XtWindow(w),IceInitiateSetupFailed);
	    major_opcode = 0;
	    return;
	}
    }
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapClntInit","major_opcode=%1d\n",major_opcode);
    InfoF(__FILE__,__LINE__,"RapClntInit","network_ids = %s\n",network_ids);
    InfoF(__FILE__,__LINE__,"RapClntInit","agentWindow = %d \n",cme->data.l[1]);
#endif

    /* Using the ICE network IDs in the ClientMessage sent by the agent
     * (Xtea gets these for us and passes them to RapClntInit), contact the 
     * agent.
     */
    ice_conn = IceOpenConnection(network_ids,
				NULL,
				False,
				major_opcode,
				255,
				error_str);
    
    if (!ice_conn) {
	XteaSendFailure(dpy,cme,XtWindow(w),IceInitiateOpenFailed);
	ErrorF(__FILE__,__LINE__,"RapClntInit",
	       "IceOpenConnection failed due to %s\n",error_str);
	return;
    }
#ifdef DEBUG
    else {
	connect_str = IceConnectionString(ice_conn);
	InfoF(__FILE__,__LINE__,"RapClntInit",
	      "IceConnectionNumber = %1d",IceConnectionNumber(ice_conn));
	InfoF(__FILE__,__LINE__,"RapClntInit",
	      "IceConnectionString = %s\n",connect_str);
	free(connect_str);
    }
#endif

    /* Establish some context for the ICE connection.
     */
    rap_info = (RapInfo)XtMalloc(sizeof(RapInfoRec));
    memset(rap_info,'\0',sizeof(RapInfoRec));
    rap_info->top_level		= w;
    rap_info->agent_window 	= cme->data.l[1];
    rap_info->ice_conn 		= ice_conn;
    rap_info->major_opcode	= major_opcode;
    
    /* Set up the protocol and use the widget as the the client data.
     */
    status = IceProtocolSetup(ice_conn,
			      major_opcode,
			      rap_info, /* client data */
			      False,
			      &major_version,
			      &minor_version,
			      &vendor_str,
			      &release_str,
			      255,
			      error_str);

    if (status != IceProtocolSetupSuccess) {
	XteaSendFailure(dpy,cme,XtWindow(w),IceInitiateSetupFailed);
	switch(status) {
	    case IceProtocolSetupFailure: {
		ErrorF(__FILE__,__LINE__,"RapClntInit",
		       "IceProtocolSetup failed due to IceProtocolSetupFailure\n");
		break;
	    }
	    case IceProtocolSetupIOError: {
		ErrorF(__FILE__,__LINE__,"RapClntInit",
		       "IceProtocolSetup failed due to IceProtocolSetupIOError\n");
		break;
	    }
	    case IceProtocolAlreadyActive: {
		ErrorF(__FILE__,__LINE__,"RapClntInit",
		       "IceProtocolSetup failed due to IceProtocolAlreadyActive\n");
		break;
	    }
	    default: {
		ErrorF(__FILE__,__LINE__,"RapClntInit",
		       "IceProtocolSetup failed.  Unknown status %d\n",status);
	    }
	}
	ErrorF(__FILE__,__LINE__,"RapClntInit",
	       "IceProtocolSetup error string: %s\n",error_str);
	XtFree((char *)rap_info);	    
    }
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapClntInit",
	  "Protocol Setup:  Version: %1d.%1d, Vendor: %s, Release %s\n",
	  major_version,minor_version,vendor_str,release_str);
#endif
	
    /* Handle ICE messages using XtAppAddInput.  Use the ice_conn 
     * connection number as a way to determine when ICE messages 
     * are being sent on the connection.  In addition, use the
     * ice_conn ID as the client data so the ProcessIceMsg can
     * handle multiple agents.
     */
    XtAppAddInput(XtWidgetToApplicationContext(w),
		  IceConnectionNumber(ice_conn),
		  (XtPointer)XtInputReadMask,
		  ProcessIceMsg,
		  (XtPointer)rap_info);

    /* Set up the handler to handle sudden breaks in the 
     * ICE connection.   [[[WDW - This really isn't such a
     * good thing to do.]]]]
     */
    IceSetErrorHandler(ErrorHandler);
    IceSetIOErrorHandler(IOErrorHandler);

    /* Add the client to the list of known clients.
     * This will be held off the extension data for the
     * display.  
     */
    RapClntAddAgent(dpy,rap_info);
        
} /* RapClntInit */


/************************************************************************/
/*									*/
/* RapClntSetup								*/
/*									*/
/*	Add the RAP protocol to the list of known protocols for the	*/
/*	client.								*/
/*									*/
/************************************************************************/
void 
RapClntSetup(Widget w)
{
    XteaAddProtocol(w,
		    XInternAtom(XtDisplay(w),ICE_INITIATE RapName,False),
		    RapClntInit,
		    NULL);

} /* RapClntSetup */
